==> morph_data-0.0.1rc7.tar.gz: morph_data-0.0.1rc7/core/morph_lib/database.py <==
from __future__ import annotations

import json
import os
from typing import Any, List, Optional, Union, get_args

import pandas as pd
import urllib3
from morph_lib.error import MorphApiError
from morph_lib.utils.db_connector import DBConnector
from morph_lib.utils.sql import SQLUtils
from pydantic import BaseModel, Field
from urllib3.exceptions import InsecureRequestWarning

from morph.api.cloud.client import MorphApiClient, MorphApiKeyClientImpl
from morph.config.project import MorphProject, load_project
from morph.task.utils.connection import (
    CONNECTION_TYPE,
    MORPH_BUILTIN_DB_CONNECTION_SLUG,
    Connection,
    ConnectionYaml,
    DatabaseConnection,
)
from morph.task.utils.connections.connector import Connector
from morph.task.utils.morph import find_project_root_dir

urllib3.disable_warnings(InsecureRequestWarning)


# ===============================================
#
# Class
#
# ===============================================
class FieldSchema(BaseModel):
    name: str
    type: str
    nullable: Optional[bool] = Field(default=True)
    primary: Optional[bool] = Field(default=False)
    comment: Optional[str] = Field(default=None)
    length: Optional[int] = Field(default=None)
    unsigned: Optional[bool] = Field(default=None)
    auto_increment: Optional[bool] = Field(default=None)
    default: Optional[Any] = Field(default=None)
    enums: Optional[List[str]] = Field(default=None)


class TableSchema(BaseModel):
    dialect: str
    name: str
    fields: List[FieldSchema]
    schema_name: Optional[str] = Field(default=None)

    def to_text(self):
        text = f"""# Database
{self.dialect}

"""
        if self.schema_name:
            text += f"""# Schema
{self.schema_name}
"""
        text += f"""# Table
{self.name}

"""
        text += "# Fields\n"
        for i, field in enumerate(self.fields):
            field_text = f"""col{i+1}. {field.name} type={field.type} nullable={field.nullable} """
            if field.primary is not None:
                field_text += f"primary={field.primary} "
            if field.comment:
                field_text += f"comment={field.comment}"
            if field.length:
                field_text += f"length={field.length}"
            if field.unsigned is not None:
                field_text += f"unsigned={field.unsigned}"
            if field.auto_increment is not None:
                field_text += f"auto_increment={field.auto_increment}"
            if field.default:
                field_text += f"default_value={field.default}"
            if field.enums:
                field_text += f"enums={field.enums}"
            text += field_text + "\n"
        return text


# ===============================================
#
# Implementation
#
# ===============================================


def __find_connection(connection: str | Connection | None) -> Connection:
    workspace_id_or_connection_slug = os.getenv("MORPH_WORKSPACE_ID", "")
    if isinstance(connection, str):
        workspace_id_or_connection_slug = connection
    if isinstance(connection, get_args(Connection)):
        return connection  # type: ignore

    cloud_connection: Optional[Union[Connection, DatabaseConnection]] = None

    if connection is not None and isinstance(connection, str):
        cloud_connection = ConnectionYaml.find_cloud_connection(connection)
        if (
            cloud_connection.type != CONNECTION_TYPE.bigquery
            and cloud_connection.type != CONNECTION_TYPE.snowflake
            and cloud_connection.type != CONNECTION_TYPE.redshift
            and cloud_connection.type != CONNECTION_TYPE.mysql
            and cloud_connection.type != CONNECTION_TYPE.postgres
            and cloud_connection.type != CONNECTION_TYPE.athena
            and cloud_connection.type != CONNECTION_TYPE.duckdb
        ):
            raise MorphApiError(
                f"Connection type {cloud_connection.type} is not supported"
            )
    else:
        project: Optional[MorphProject] = load_project(find_project_root_dir())
        if project and project.default_connection:
            default_connection = project.default_connection
            connection_yaml = ConnectionYaml.load_yaml()
            if default_connection == MORPH_BUILTIN_DB_CONNECTION_SLUG:
                cloud_connection = ConnectionYaml.find_connection(
                    connection_yaml, workspace_id_or_connection_slug
                )
                if cloud_connection is None:
                    db, cloud_connection = ConnectionYaml.find_builtin_db_connection()
                    connection_yaml.add_connections({db: cloud_connection})
                    connection_yaml.save_yaml(True)
            else:
                cloud_connection = ConnectionYaml.find_connection(
                    connection_yaml, default_connection
                )
                if cloud_connection is None:
                    cloud_connection = ConnectionYaml.find_cloud_connection(
                        default_connection
                    )
        else:
            connection_yaml = ConnectionYaml.load_yaml()
            cloud_connection = ConnectionYaml.find_connection(
                connection_yaml, workspace_id_or_connection_slug
            )
            if cloud_connection is None:
                db, cloud_connection = ConnectionYaml.find_builtin_db_connection()
                connection_yaml.add_connections({db: cloud_connection})
                connection_yaml.save_yaml(True)

    return cloud_connection


def __execute_sql_impl(
    sql: str,
    connection: str | Connection | None = None,
) -> pd.DataFrame:
    cloud_connection: Connection = __find_connection(connection)

    connector = Connector(
        connection if isinstance(connection, str) else "",
        cloud_connection,
        is_cloud=True,
    )
    return connector.execute_sql(sql)


def __process_records(
    action: str,
    data: pd.DataFrame,
    table_name: str,
    primary_keys: List[str] = [],
    connection: Optional[str] = None,
) -> None:
    # Validate action
    if action not in {"insert", "update", "delete", "insert_or_update"}:
        raise MorphApiError(
            "Invalid action provided. Must be 'create', 'insert', or 'update'."
        )

    cloud_connection = __find_connection(connection)
    if (
        cloud_connection.type != CONNECTION_TYPE.postgres
        and cloud_connection.type != CONNECTION_TYPE.snowflake
        and cloud_connection.type != CONNECTION_TYPE.redshift
        and cloud_connection.type != CONNECTION_TYPE.mysql
        and cloud_connection.type != CONNECTION_TYPE.bigquery
    ):
        raise MorphApiError(
            "Only PostgreSQL, Snowflake, Redshift, BigQuery, and MySQL are supported for now."
        )

    if action == "update" or action == "delete" or action == "insert_or_update":
        if len(primary_keys) == 0:
            raise MorphApiError(
                "Primary keys are required for update, delete, and insert_or_update actions."
            )
        missing_keys = [key for key in primary_keys if key not in data.columns]
        if len(missing_keys) > 0:
            raise MorphApiError(
                f"Primary keys {missing_keys} are not present in the DataFrame columns."
            )

    sql_utils = SQLUtils(data, table_name, cloud_connection)

    if action == "insert":
        __execute_sql_impl(sql_utils.generate_insert_sql(), connection=cloud_connection)
    elif action == "update":
        __execute_sql_impl(
            sql_utils.generate_update_sql(primary_keys), connection=cloud_connection
        )
    elif action == "insert_or_update":
        if (
            cloud_connection.type == CONNECTION_TYPE.redshift
            or cloud_connection.type == CONNECTION_TYPE.bigquery
        ):
            raise MorphApiError("Redshift and BigQuery do not support UPSERT.")
        __execute_sql_impl(
            sql_utils.generate_insert_or_update_sql(primary_keys),
            connection=cloud_connection,
        )
    elif action == "delete":
        __execute_sql_impl(
            sql_utils.generate_delete_sql(primary_keys), connection=cloud_connection
        )


# ===============================================
#
# Functions
#
# ===============================================
def execute_sql(sql: str, connection: Optional[str] = None) -> pd.DataFrame:
    """
    Execute SQL query
    """
    return __execute_sql_impl(sql, connection)


def insert_records(
    data: pd.DataFrame,
    table_name: str,
    connection: Optional[str] = None,
) -> None:
    """
    Insert records into the table
    """
    __process_records("insert", data, table_name, connection=connection)


def update_records(
    data: pd.DataFrame,
    primary_keys: List[str],
    table_name: str,
    connection: Optional[str] = None,
) -> None:
    """
    Update records in the table
    """
    __process_records(
        "update", data, table_name, primary_keys=primary_keys, connection=connection
    )


def insert_or_update_records(
    data: pd.DataFrame,
    primary_keys: List[str],
    table_name: str,
    connection: Optional[str] = None,
) -> None:
    """
    Insert or Update records in the table
    """
    __process_records(
        "insert_or_update",
        data,
        table_name,
        primary_keys=primary_keys,
        connection=connection,
    )


def delete_records(
    data: pd.DataFrame,
    primary_keys: List[str],
    table_name: str,
    connection: Optional[str] = None,
) -> None:
    """
    Delete records in the table
    """
    __process_records(
        "delete", data, table_name, primary_keys=primary_keys, connection=connection
    )


def get_db_connector(connection_slug: Optional[str] = None) -> DBConnector:
    """
    Obtain a DBConnector object for PostgreSQL, Redshift, or MySQL
    """
    cloud_connection: Optional[Union[Connection, DatabaseConnection]] = None

    if connection_slug and connection_slug != MORPH_BUILTIN_DB_CONNECTION_SLUG:
        connection_yaml = ConnectionYaml.load_yaml()
        cloud_connection = ConnectionYaml.find_connection(
            connection_yaml, connection_slug
        )
        if cloud_connection is None:
            cloud_connection = ConnectionYaml.find_cloud_connection(connection_slug)
        if (
            cloud_connection.type != CONNECTION_TYPE.postgres
            and cloud_connection.type != CONNECTION_TYPE.redshift
            and cloud_connection.type != CONNECTION_TYPE.mysql
        ):
            raise ValueError(f"Connection '{connection_slug}' is not supported.")
    else:
        _, cloud_connection = ConnectionYaml.find_builtin_db_connection()

    return DBConnector(cloud_connection)


def get_tables(
    table_names: List[str],
    schema_name: Optional[str],
    connection: Optional[str],
) -> List[TableSchema]:
    """
    Get the table structures
    """
    connection_ = __find_connection(connection)

    tables: List[TableSchema] = []
    client = MorphApiClient(MorphApiKeyClientImpl)
    for table in table_names:
        response = client.req.list_fields(table, schema_name, connection)
        try:
            response.is_error(True)
        except Exception as e:
            error_detail = {
                "type": type(e).__name__,
                "message": str(e),
            }
            error_json = json.dumps(error_detail, ensure_ascii=False)
            raise MorphApiError(error_json)

        fields: List[FieldSchema] = []
        for field in response.json()["fields"]:
            field_type = field["nativeType"] if "nativeType" in field else field["type"]
            fields.append(
                FieldSchema(
                    name=field["name"],
                    type=field_type,
                    nullable=field["nullable"] if "nullable" in field else True,
                    primary=field["primary"] if "primary" in field else None,
                    comment=field["comment"] if "comment" in field else None,
                    length=field["length"] if "length" in field else None,
                    unsigned=field["unsigned"] if "unsigned" in field else None,
                    auto_increment=(
                        field["autoIncrement"] if "autoIncrement" in field else None
                    ),
                    default=(field["default"] if "default" in field else None),
                    enums=field["members"] if "members" in field else None,
                )
            )

        tables.append(
            TableSchema(
                dialect=connection_.type,
                name=table,
                fields=fields,
                schema_name=schema_name,
            )
        )

    return tables

==> morph_data-0.0.3.tar.gz: morph_data-0.0.3/core/morph_lib/database.py <==
from __future__ import annotations

import json
import os
from typing import Any, List, Optional, Union, get_args

import pandas as pd
import urllib3
from morph_lib.error import MorphApiError
from morph_lib.utils.db_connector import DBConnector
from morph_lib.utils.sql import SQLUtils
from pydantic import BaseModel, Field
from urllib3.exceptions import InsecureRequestWarning

from morph.api.cloud.client import MorphApiClient, MorphApiKeyClientImpl
from morph.api.cloud.utils import is_cloud
from morph.config.project import MorphProject, load_project
from morph.task.utils.connection import (
    CONNECTION_TYPE,
    MORPH_BUILTIN_DB_CONNECTION_SLUG,
    Connection,
    ConnectionYaml,
    DatabaseConnection,
    DuckDBConnection,
)
from morph.task.utils.connections.connector import Connector
from morph.task.utils.morph import find_project_root_dir

urllib3.disable_warnings(InsecureRequestWarning)


# ===============================================
#
# Class
#
# ===============================================
class FieldSchema(BaseModel):
    name: str
    type: str
    nullable: Optional[bool] = Field(default=True)
    primary: Optional[bool] = Field(default=False)
    comment: Optional[str] = Field(default=None)
    length: Optional[int] = Field(default=None)
    unsigned: Optional[bool] = Field(default=None)
    auto_increment: Optional[bool] = Field(default=None)
    default: Optional[Any] = Field(default=None)
    enums: Optional[List[str]] = Field(default=None)


class TableSchema(BaseModel):
    dialect: str
    name: str
    fields: List[FieldSchema]
    schema_name: Optional[str] = Field(default=None)

    def to_text(self):
        text = f"""# Database
{self.dialect}

"""
        if self.schema_name:
            text += f"""# Schema
{self.schema_name}
"""
        text += f"""# Table
{self.name}

"""
        text += "# Fields\n"
        for i, field in enumerate(self.fields):
            field_text = f"""col{i+1}. {field.name} type={field.type} nullable={field.nullable} """
            if field.primary is not None:
                field_text += f"primary={field.primary} "
            if field.comment:
                field_text += f"comment={field.comment}"
            if field.length:
                field_text += f"length={field.length}"
            if field.unsigned is not None:
                field_text += f"unsigned={field.unsigned}"
            if field.auto_increment is not None:
                field_text += f"auto_increment={field.auto_increment}"
            if field.default:
                field_text += f"default_value={field.default}"
            if field.enums:
                field_text += f"enums={field.enums}"
            text += field_text + "\n"
        return text


# ===============================================
#
# Implementation
#
# ===============================================


def __find_connection(connection: str | Connection | None) -> Connection:
    workspace_id_or_connection_slug = os.getenv("MORPH_WORKSPACE_ID", "")
    if isinstance(connection, str):
        workspace_id_or_connection_slug = connection
    if isinstance(connection, get_args(Connection)):
        return connection  # type: ignore

    cloud_connection: Optional[Union[Connection, DatabaseConnection]] = None

    if connection is not None and isinstance(connection, DuckDBConnection):
        return connection
    elif connection is not None and isinstance(connection, str):
        if not is_cloud():
            connection_yaml = ConnectionYaml.load_yaml()
            cloud_connection = ConnectionYaml.find_connection(
                connection_yaml, connection
            )
            if cloud_connection is None:
                raise MorphApiError(f"Could not find {connection} in connections.yml.")
            return cloud_connection
        cloud_connection = ConnectionYaml.find_cloud_connection(connection)
        if (
            cloud_connection.type != CONNECTION_TYPE.bigquery
            and cloud_connection.type != CONNECTION_TYPE.snowflake
            and cloud_connection.type != CONNECTION_TYPE.redshift
            and cloud_connection.type != CONNECTION_TYPE.mysql
            and cloud_connection.type != CONNECTION_TYPE.postgres
            and cloud_connection.type != CONNECTION_TYPE.athena
            and cloud_connection.type != CONNECTION_TYPE.duckdb
        ):
            raise MorphApiError(
                f"Connection type {cloud_connection.type} is not supported"
            )
    else:
        project: Optional[MorphProject] = load_project(find_project_root_dir())
        if project and project.default_connection:
            default_connection = project.default_connection
            connection_yaml = ConnectionYaml.load_yaml()
            if default_connection == MORPH_BUILTIN_DB_CONNECTION_SLUG:
                cloud_connection = ConnectionYaml.find_connection(
                    connection_yaml, workspace_id_or_connection_slug
                )
                if cloud_connection is None:
                    db, cloud_connection = ConnectionYaml.find_builtin_db_connection()
                    connection_yaml.add_connections({db: cloud_connection})
                    connection_yaml.save_yaml(True)
            else:
                cloud_connection = ConnectionYaml.find_connection(
                    connection_yaml, default_connection
                )
                if cloud_connection is None:
                    cloud_connection = ConnectionYaml.find_cloud_connection(
                        default_connection
                    )
        else:
            connection_yaml = ConnectionYaml.load_yaml()
            cloud_connection = ConnectionYaml.find_connection(
                connection_yaml, workspace_id_or_connection_slug
            )
            if cloud_connection is None:
                db, cloud_connection = ConnectionYaml.find_builtin_db_connection()
                connection_yaml.add_connections({db: cloud_connection})
                connection_yaml.save_yaml(True)

    return cloud_connection


def __execute_sql_impl(
    sql: str,
    connection: str | Connection | None = None,
) -> pd.DataFrame:
    cloud_connection: Connection = __find_connection(connection)

    connector = Connector(
        connection if isinstance(connection, str) else "",
        cloud_connection,
        is_cloud=True,
    )
    return connector.execute_sql(sql)


def __process_records(
    action: str,
    data: pd.DataFrame,
    table_name: str,
    primary_keys: List[str] = [],
    connection: Optional[str] = None,
) -> None:
    # Validate action
    if action not in {"insert", "update", "delete", "insert_or_update"}:
        raise MorphApiError(
            "Invalid action provided. Must be 'create', 'insert', or 'update'."
        )

    cloud_connection = __find_connection(connection)
    if (
        cloud_connection.type != CONNECTION_TYPE.postgres
        and cloud_connection.type != CONNECTION_TYPE.snowflake
        and cloud_connection.type != CONNECTION_TYPE.redshift
        and cloud_connection.type != CONNECTION_TYPE.mysql
        and cloud_connection.type != CONNECTION_TYPE.bigquery
    ):
        raise MorphApiError(
            "Only PostgreSQL, Snowflake, Redshift, BigQuery, and MySQL are supported for now."
        )

    if action == "update" or action == "delete" or action == "insert_or_update":
        if len(primary_keys) == 0:
            raise MorphApiError(
                "Primary keys are required for update, delete, and insert_or_update actions."
            )
        missing_keys = [key for key in primary_keys if key not in data.columns]
        if len(missing_keys) > 0:
            raise MorphApiError(
                f"Primary keys {missing_keys} are not present in the DataFrame columns."
            )

    sql_utils = SQLUtils(data, table_name, cloud_connection)

    if action == "insert":
        __execute_sql_impl(sql_utils.generate_insert_sql(), connection=cloud_connection)
    elif action == "update":
        __execute_sql_impl(
            sql_utils.generate_update_sql(primary_keys), connection=cloud_connection
        )
    elif action == "insert_or_update":
        if (
            cloud_connection.type == CONNECTION_TYPE.redshift
            or cloud_connection.type == CONNECTION_TYPE.bigquery
        ):
            raise MorphApiError("Redshift and BigQuery do not support UPSERT.")
        __execute_sql_impl(
            sql_utils.generate_insert_or_update_sql(primary_keys),
            connection=cloud_connection,
        )
    elif action == "delete":
        __execute_sql_impl(
            sql_utils.generate_delete_sql(primary_keys), connection=cloud_connection
        )


# ===============================================
#
# Functions
#
# ===============================================
def execute_sql(sql: str, connection: Optional[str] = None) -> pd.DataFrame:
    """
    Execute SQL query
    """
    return __execute_sql_impl(sql, connection)


def insert_records(
    data: pd.DataFrame,
    table_name: str,
    connection: Optional[str] = None,
) -> None:
    """
    Insert records into the table
    """
    __process_records("insert", data, table_name, connection=connection)


def update_records(
    data: pd.DataFrame,
    primary_keys: List[str],
    table_name: str,
    connection: Optional[str] = None,
) -> None:
    """
    Update records in the table
    """
    __process_records(
        "update", data, table_name, primary_keys=primary_keys, connection=connection
    )


def insert_or_update_records(
    data: pd.DataFrame,
    primary_keys: List[str],
    table_name: str,
    connection: Optional[str] = None,
) -> None:
    """
    Insert or Update records in the table
    """
    __process_records(
        "insert_or_update",
        data,
        table_name,
        primary_keys=primary_keys,
        connection=connection,
    )


def delete_records(
    data: pd.DataFrame,
    primary_keys: List[str],
    table_name: str,
    connection: Optional[str] = None,
) -> None:
    """
    Delete records in the table
    """
    __process_records(
        "delete", data, table_name, primary_keys=primary_keys, connection=connection
    )


def get_db_connector(connection_slug: Optional[str] = None) -> DBConnector:
    """
    Obtain a DBConnector object for PostgreSQL, Redshift, or MySQL
    """
    cloud_connection: Optional[Union[Connection, DatabaseConnection]] = None

    if connection_slug and connection_slug != MORPH_BUILTIN_DB_CONNECTION_SLUG:
        connection_yaml = ConnectionYaml.load_yaml()
        cloud_connection = ConnectionYaml.find_connection(
            connection_yaml, connection_slug
        )
        if cloud_connection is None:
            cloud_connection = ConnectionYaml.find_cloud_connection(connection_slug)
        if (
            cloud_connection.type != CONNECTION_TYPE.postgres
            and cloud_connection.type != CONNECTION_TYPE.redshift
            and cloud_connection.type != CONNECTION_TYPE.mysql
        ):
            raise ValueError(f"Connection '{connection_slug}' is not supported.")
    else:
        _, cloud_connection = ConnectionYaml.find_builtin_db_connection()

    return DBConnector(cloud_connection)


def get_tables(
    table_names: List[str],
    schema_name: Optional[str],
    connection: Optional[str],
) -> List[TableSchema]:
    """
    Get the table structures
    """
    connection_ = __find_connection(connection)

    tables: List[TableSchema] = []
    client = MorphApiClient(MorphApiKeyClientImpl)
    for table in table_names:
        response = client.req.list_fields(table, schema_name, connection)
        try:
            response.is_error(True)
        except Exception as e:
            error_detail = {
                "type": type(e).__name__,
                "message": str(e),
            }
            error_json = json.dumps(error_detail, ensure_ascii=False)
            raise MorphApiError(error_json)

        fields: List[FieldSchema] = []
        for field in response.json()["fields"]:
            field_type = field["nativeType"] if "nativeType" in field else field["type"]
            fields.append(
                FieldSchema(
                    name=field["name"],
                    type=field_type,
                    nullable=field["nullable"] if "nullable" in field else True,
                    primary=field["primary"] if "primary" in field else None,
                    comment=field["comment"] if "comment" in field else None,
                    length=field["length"] if "length" in field else None,
                    unsigned=field["unsigned"] if "unsigned" in field else None,
                    auto_increment=(
                        field["autoIncrement"] if "autoIncrement" in field else None
                    ),
                    default=(field["default"] if "default" in field else None),
                    enums=field["members"] if "members" in field else None,
                )
            )

        tables.append(
            TableSchema(
                dialect=connection_.type,
                name=table,
                fields=fields,
                schema_name=schema_name,
            )
        )

    return tables
